#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  3 11:38:52 2019

# D McPhillips August 2019
"""

from PyQt5.QtWidgets import QMainWindow
import dpeUi_v1p5 
import numpy as np
from scipy.interpolate import interp2d
import pandas as pd
from matplotlib.path import Path
from matplotlib.widgets import LassoSelector
import matplotlib.colors as mcrls
import glob as gb

class MyApp(QMainWindow, dpeUi_v1p5.Ui_MplMainWindow):
    def __init__(self):
        super(self.__class__, self).__init__()
        self.setupUi(self)
        
        self.kdedata = np.array([[np.nan],[np.nan]]) 
        self.otherdata = np.array([np.nan])
        self.kdedata_old = None 
        self.otherdata_old = None
        self.ndata = list([0]) 
        self.fnamestor = list('') 
        self.toggleAxisFreq = True 
        self.toggleOutPolyDel = True
        self.toggleOutPolyfit = True
        self.outsumm = None
        self.indp = None
        self.lasso = None
        self.selectFlag = False
        self.id = np.array([])
        self.id_old = None
        self.fH = None
        self.xedge = None
        self.yedge = None
        self.origdata = None
        
        self.xaxis_tog.setChecked(True)
        self.xaxis_tog.stateChanged.connect(self.toggleAxis)
        
        self.outPolyDel_tog.setChecked(True)
        self.outPolyDel_tog.stateChanged.connect(self.togglePolyDel)
        
        self.polyfit_tog.setChecked(True)
        self.polyfit_tog.stateChanged.connect(self.togglePolyfit)
        
        self.loadPlot_btn.clicked.connect(self.loadData)
        
        self.rePlot_btn.clicked.connect(self.plotData)
        
        self.selectPts_btn.clicked.connect(self.selectPts)
        
        self.cancel_btn.clicked.connect(self.cancelPts)
        
        self.acceptPts_btn.clicked.connect(self.acceptPts)
        self.acceptPts_btn.clicked.connect(self.plotData)
        
        self.undo_btn.clicked.connect(self.undo)
        self.undo_btn.clicked.connect(self.plotData)
        
        self.dataSumm_btn.clicked.connect(self.summData)
        
        self.save_btn.clicked.connect(self.saveData)
        
        self.nbin_bx.setValue(18)
        self.nLog_txt.setText('0.5')
        self.overLap_bx.setValue(3)
        self.fileIn_txt.setPlainText('(e.g., data/filename)') 

    def loadData(self):
        # Load data. Repeated clicks load new files but do not reload existing ones
        self.fnamestor = list('')
        self.ndata = list([0])
        self.kdedata = np.array([[np.nan],[np.nan]])
        self.otherdata = np.array([np.nan])
        self.id = np.array([])
        
        fnameinX = self.fileIn_txt.toPlainText()
        nl = fnameinX.find('\n') 
        if nl > -1: 
            nlines = 1 
            for i in fnameinX:
                nlines = nlines + int(i=='\n')
            for i in range(nlines):
                nl = fnameinX.find('\n')
                if nl == -1:
                    fnamein = fnameinX
                else:
                    fnamein = fnameinX[:nl]
                if len(fnamein) != 0: 
                    MyApp.subloadData(fnamein,self)
                    fnameinX = fnameinX[(nl+1):]
                
        else:
            MyApp.subloadData(fnameinX,self)

        if self.toggleAxisFreq:
            self.origdata = np.vstack([self.kdedata[1,:],
                                      self.kdedata[0,:],self.otherdata])
            self.fH,self.xedge,self.yedge = (np.histogram2d(self.origdata[1,:],
                                             self.origdata[0,:],bins
                                             =[np.logspace(np.log10(
                                             min(self.origdata[1,:])),
                                             np.log10(max(self.origdata[1,:])),64),
                                             np.linspace(min(self.origdata[0,:]),
                                             max(self.origdata[0,:]),64)]))
            self.fH = self.fH.T
        else:
            self.origdata = (np.vstack([self.kdedata[0,:],
                             self.otherdata,self.kdedata[1,:]]))
            self.fH,self.xedge,self.yedge = (np.histogram2d(self.origdata[0,:],
                                             self.origdata[2,:],bins=[np.linspace(
                                             min(self.origdata[0,:]),
                                             max(self.origdata[0,:]),64),
                                             np.logspace(np.log10(
                                             min(self.origdata[2,:])),
                                             np.log10(max(self.origdata[2,:])),64)]))
            self.fH = self.fH.T
        
    def subloadData(fnamein,self): 
        #Subroutine for using glob tools for loading data
        for f in gb.glob(fnamein):
            df = pd.read_csv(f,sep = '\s+|\t| |,',engine = 'python')
            
            self.fnamestor.extend([f])

            df.replace([np.inf, -np.inf],np.nan)
            df.dropna() 
            
            self.ndata.append(len(df.Velocity)) 
            if df.shape[1] > 2:
                if self.toggleAxisFreq:
                    self.kdedata = (np.hstack((self.kdedata,
                                    np.vstack([df.Frequency,df.Velocity]))))
                    self.otherdata = np.hstack((self.otherdata,df.Wavelength))
                else:
                    self.kdedata = np.vstack([df.Velocity,df.Wavelength])
                    self.otherdata = df.Frequency
            else:
                if self.toggleAxisFreq:
                    self.kdedata = (np.hstack((self.kdedata,
                                    np.vstack([df.Frequency,df.Velocity]))))
                    self.otherdata = np.ones(len(self.kdedata))
                else:
                    self.kdedata = np.vstack([df.Velocity,df.Wavelength])
                    self.otherdata = np.ones(len(self.kdedata))
            self.ndata[0] = self.ndata[0] + 1.0
            self.id = np.hstack((self.id,(self.ndata[0]-1)*np.ones(self.ndata[-1]))) 
            
            if len(self.ndata) == 2: 
                self.kdedata = self.kdedata[:,1:] 
                self.otherdata = self.otherdata[1:]
        
        fout1 = ''
        fout2 = ''
        for x in self.fnamestor:
            fout1 = fout1 + x + '\n'
            fout2 = fout2 + '1' + '\n'
        fout1 = fout1[:-1]
        fout2 = fout2[:-1]
        self.datasetOut_txt.setPlainText(fout1) 
        self.datasetW_txt.setPlainText(fout2) 
        

    def toggleAxis(self):
        #Change the variables
        self.toggleAxisFreq = np.invert(self.toggleAxisFreq)
        if not np.isnan(self.kdedata[0]): #only manipulate if data is loaded
            if self.toggleAxisFreq:
                temp = self.otherdata
                self.otherdata = self.kdedata[1,:]
                self.kdedata = np.vstack([temp,self.kdedata[0,:]])
                self.fH,self.xedge,self.yedge = (np.histogram2d(self.origdata[1,:],
                                                 self.origdata[0,:],bins=[np.logspace(
                                                 np.log10(min(self.origdata[1,:])),
                                                 np.log10(max(self.origdata[1,:])),64),
                                                 np.linspace(min(self.origdata[0,:]),
                                                 max(self.origdata[0,:]),64)])) 
                self.fH = self.fH.T
            else:
                temp = self.otherdata
                self.otherdata = self.kdedata[0,:]
                self.kdedata = np.vstack([self.kdedata[1,:],temp])
                self.fH,self.xedge,self.yedge = (np.histogram2d(self.origdata[0,:],
                                                 self.origdata[2,:],bins=[np.linspace(
                                                 min(self.origdata[0,:]),
                                                 max(self.origdata[0,:]),64),
                                                 np.logspace(np.log10(
                                                 min(self.origdata[2,:])),
                                                 np.log10(max(self.origdata[2,:])),64)])) 
                self.fH = self.fH.T
        
    def togglePolyDel(self):
        self.toggleOutPolyDel = np.invert(self.toggleOutPolyDel)
        
    def togglePolyfit(self):
        self.toggleOutPolyfit = np.invert(self.toggleOutPolyfit)

    def plotData(self):
        #Plot data grouped in log-width bins
        nlogBase = float(self.nLog_txt.text()) #user inputs number of bins
        Z = self.fH
        fZ = (interp2d(self.xedge[:-1]+(self.xedge[2]-self.xedge[1])/2,self.yedge[:-1]
              +(self.yedge[2]-self.yedge[1])/2,Z,kind='cubic'))
        if self.toggleAxisFreq:
            _, xdg, ydg = (np.histogram2d(self.kdedata[0,:],self.kdedata[1,:],
                           bins=[np.logspace(np.log10(min(self.kdedata[0,:])),
                           np.log10(max(self.kdedata[0,:])),64),np.linspace(
                           min(self.kdedata[1,:]),
                           max(self.kdedata[1,:]),64)]))
            xe2 = np.logspace(np.log10(min(xdg)),np.log10(max(xdg)),96)
            ye2 = np.linspace(min(ydg),max(ydg),96)
        else:
            _, xdg, ydg = (np.histogram2d(self.kdedata[0,:],self.kdedata[1,:],
                           bins=[np.linspace(min(self.kdedata[0,:]),
                           max(self.kdedata[0,:]),64),np.logspace(np.log10(
                           min(self.kdedata[1,:])),
                           np.log10(max(self.kdedata[1,:])),64)]))
            xe2 = np.linspace(min(xdg),max(xdg),96)
            ye2 = np.logspace(np.log10(min(ydg)),np.log10(max(ydg)),96)
        X2,Y2 = np.meshgrid(xe2,ye2)
        Z = fZ(xe2,ye2)
        Z[Z<=0] = 0 
        
        self.mpl.canvas.ax.cla()   
        if self.toggleAxisFreq:
            self.mpl.canvas.ax.pcolormesh(X2,Y2,Z,cmap='coolwarm',shading = 'gourand',
                                          norm=mcrls.PowerNorm(gamma=nlogBase))
            itot = 0
            ic = 0
            colors = np.linspace(0.01,0.89,self.ndata[0]) #create list for grayscale values
            for i in self.ndata[1:]:
                itot = itot + i
                self.mpl.canvas.ax.plot(self.kdedata[0,(itot-i):(itot-1)],
                                        self.kdedata[1,(itot-i):(itot-1)],
                                        linestyle = 'none',color=str(colors[ic]),
                                        marker='o',markersize = 0.5)
                ic = ic + 1
            self.mpl.canvas.ax.set_xscale('log')
            self.mpl.canvas.ax.set_xlabel('Frequency')
            self.mpl.canvas.ax.set_ylabel('Velocity')
        else:
            self.mpl.canvas.ax.pcolormesh(X2,Y2,Z,cmap='coolwarm',
                                          shading = 'gourand',
                                          norm=mcrls.PowerNorm(gamma=nlogBase))
            itot = 0
            ic = 0
            colors = np.linspace(0.01,0.89,self.ndata[0])
            for i in self.ndata[1:]:
                itot = itot + i
                self.mpl.canvas.ax.plot(self.kdedata[0,(itot-i):(itot-1)],
                                        self.kdedata[1,(itot-i):(itot-1)],
                                        linestyle = 'none',color=str(colors[ic]),
                                        marker='o',markersize = 0.5)
                ic = ic + 1
            self.mpl.canvas.ax.set_yscale('log')
            self.mpl.canvas.ax.set_xlabel('Velocity')
            self.mpl.canvas.ax.set_ylabel('Wavelength')
            self.mpl.canvas.ax.invert_yaxis() 
        self.mpl.canvas.draw()
    
    def selectPts(self):
        #Select and delete undesireable data
        self.lasso = LassoSelector(self.mpl.canvas.ax, onselect = self.onselect)
        
    def onselect(self,verts):
        path = Path(verts)
        self.indp = path.contains_points(self.kdedata.T)
        self.selectFlag = True
        
    def cancelPts(self):
        if self.selectFlag:
            self.lasso.disconnect_events()
            self.selectFlag = False
        
    def acceptPts(self):
        self.lasso.disconnect_events()
        if self.selectFlag:
            if self.toggleOutPolyDel:
                kdedata_new = self.kdedata.T[self.indp].T
                otherdata_new = self.otherdata.T[self.indp].T 
                id_new = self.id.T[self.indp].T
            else:
                kdedata_new = self.kdedata.T[np.invert(self.indp)].T 
                otherdata_new = self.otherdata.T[np.invert(self.indp)].T 
                id_new = self.id.T[np.invert(self.indp)].T
            
            self.kdedata_old = self.kdedata
            self.otherdata_old = self.otherdata
            self.id_old = self.id
            self.kdedata = kdedata_new
            self.otherdata = otherdata_new
            self.id = id_new
            self.selectFlag = False
            
            for i in range(int(self.ndata[0])):
                j = i+1
                self.ndata[j] = sum(self.id==i) 
    
    def undo(self):
        self.kdedata = self.kdedata_old
        self.otherdata = self.otherdata_old
        self.id = self.id_old
        for i in range(int(self.ndata[0])):
            j = i+1
            self.ndata[j] = sum(self.id==i)
    
    def summData(self):
        # Summarize the trimmed data. Use piecewise polynomial after Joh(1996)
        # Weight the data in two ways (1) by inverse number of points in the
        # bin, and (2) by user-defined relative weight. Alternately, use simple
        # median summary if desired by the user
        
        nbins = self.nbin_bx.value() 
        wuin = self.datasetW_txt.toPlainText() 
        noverLap = self.overLap_bx.value() 
        nlines = 1 
        
        if nbins <= 1:
           print('Must use more the one bin (reset)')
           self.nbin_bx.setValue(18)
           nbins = 18
        if noverLap < 1:
           print('Must use at least span of 1 bin')
           self.overLap_bx.setValue(2)
           noverLap = 2
        if nbins<(2*noverLap):
            #can't have fewer bins than overlap
            print('Bad user input halfwidth (reset)')
            self.nbin_bx.setValue(int(2*noverLap))
            nbins = 2*noverLap
            
        for i in wuin: 
            nlines = nlines + int(i=='\n')
        if nlines != int(self.ndata[0]):
                print('Bad user input weights (reset)')
                fout2 = ''
                for i in range(int(self.ndata[0])):
                    fout2 = fout2 + '1' + '\n'
                fout2 = fout2[:-1] 
                self.datasetW_txt.setPlainText(fout2) 
                wuin = self.datasetW_txt.toPlainText() 
                nlines = 1 
                for i in wuin: 
                    nlines = nlines + int(i=='\n')
                       
        wu = np.ones(int(self.ndata[0]))
        for i in range(nlines):
            nl = wuin.find('\n')
            if nl == -1:
                wu[i] = float(wuin)
            else:
                wu[i] = float(wuin[:nl])
                wuin = wuin[nl+1:]
        if sum(wu)>self.ndata[0]:
            print('Weights must be on the interval [0,1] (reset)')
            wu = np.ones(int(self.ndata[0]))
            fout2 = ''
            for i in range(int(self.ndata[0])):
                fout2 = fout2 + '1' + '\n'
            fout2 = fout2[:-1] 
            self.datasetW_txt.setPlainText(fout2)
        
        ebins = np.exp(np.histogram(np.log(self.kdedata[0,:]), bins=nbins)[1]) 
        self.outsumm = np.ones([3,nbins]) 
        nn = len(self.kdedata[0,:])
            
        for i in range(nbins): 
            if self.toggleOutPolyfit: 
                indall = np.zeros(nn, dtype=bool)
                indbin = np.zeros(nn, dtype=bool)
                for j in range(nn): 
                    if i<noverLap:
                        sde = noverLap - i
                        nlpL = noverLap - sde
                        nlpR = noverLap + sde
                    elif i>(nbins-noverLap):
                        sde = i - (nbins-noverLap)
                        nlpL = noverLap + sde
                        nlpR = noverLap - sde
                    else:
                        nlpL = noverLap
                        nlpR = noverLap
                    ind1 = self.kdedata[0,j]>ebins[i-nlpL]
                    ind2 = self.kdedata[0,j]<ebins[i+nlpR]
                    indall[j] =  ind1 & ind2
                    ind1 = self.kdedata[0,j]>ebins[i] 
                    ind2 = self.kdedata[0,j]<ebins[i+1]
                    indbin[j] = ind1 & ind2
                wc = np.ones(len(self.kdedata[1,indall])) 
                for k in range(int(self.ndata[0])):
                    temp = self.id[indall]==k 
                    if sum(temp)>0:
                        wc[temp] = 1/(sum(temp))*wu[k] 
                xmt = ebins[i] + 0.5*(ebins[i+1]-ebins[i]) 
                b0 = np.polyfit(self.kdedata[0,indall],self.kdedata[1,indall],2,w=wc) 
                ymt = b0[0]*xmt**2 + b0[1]*xmt + b0[2] 
                ysd = 2*np.std(self.kdedata[1,indbin])
                if i == 0:
                    xmt = ebins[i] + 0.01*ebins[i]
                    ymt = b0[0]*xmt**2 + b0[1]*xmt + b0[2] 
                elif i == (nbins-1):
                    xmt = ebins[nbins] - 0.01*ebins[nbins]
                    ymt = b0[0]*xmt**2 + b0[1]*xmt + b0[2] 
            else: 
                indall = np.zeros(nn, dtype=bool)
                for j in range(nn): 
                    ind1 = self.kdedata[0,j]>ebins[i]
                    ind2 = self.kdedata[0,j]<ebins[i+1]
                    indall[j] =  ind1 & ind2
                    #now get weights per dataset
                wc = np.ones(len(self.kdedata[1,indall])) 
                for k in range(int(self.ndata[0])):
                    temp = self.id[indall]==k 
                    if sum(temp)>0:
                        wc[temp] = 1/(sum(temp))*wu[k] 
                xmt = ebins[i] + 0.5*(ebins[i+1]-ebins[i]) 
                ymt = np.average(self.kdedata[1,indall],weights=wc)
                ysd = 2*np.std(self.kdedata[1,indall])
            self.outsumm[:,i] = [(xmt),(ymt),(ysd)]
            
        self.mpl.canvas.ax.errorbar(self.outsumm[0,:],self.outsumm[1,:],
                                    yerr = self.outsumm[2,:], markerfacecolor = 'g',
                                    marker = '.', linestyle = '',
                                    markeredgecolor='w', markersize = 10)
        self.mpl.canvas.draw()

    def saveData(self):
        # Save the results as CSV files
        
        temp = np.diff(self.id)
        if np.max(temp) > 1: #prevent bug from creating misleading saved data
            print(' - Error: Input dataset deleted entirely. Ouput not permitted')
            return None
        
        fnameout = self.fileOutLn_txt.text()
        if self.toggleAxisFreq:
            for i in range(int(self.ndata[0])):
                indk = self.id == i    
                dfout = pd.DataFrame({'Velocity':self.kdedata[1,indk].T,
                                     'Frequency':self.kdedata[0,indk].T,
                                     'Wavelength':self.otherdata[indk].T})
                tempout = self.fnamestor[i]
                tempout = tempout.replace('*','star')
                tempout = tempout.replace('.csv','')
                tempout = tempout.replace('.txt','')
                tempout = tempout + '_trim' + '.csv'
                dfout.to_csv(tempout,sep = '\t',index=False)
            if len(fnameout) > 1:
            	if self.outsumm is not None:
                	dfout_summ = pd.DataFrame({'Velocity':self.outsumm[1,:].T,
                	                          'Frequency':self.outsumm[0,:].T,
                	                          'Wavelength':np.ones(len(
                	                          self.outsumm[1,:])).T,
                	                          '2StDev':self.outsumm[2,:].T})
                	dfout_summ.to_csv(fnameout,sep = '\t',index=False)
        else:
            for i in range(int(self.ndata[0])):
                indk = self.id == i  
                dfout = pd.DataFrame({'Velocity':self.kdedata[0,indk].T,
                                     'Wavelength':self.kdedata[1,indk].T,
                                     'Frequency':self.otherdata[indk].T})
                tempout = self.fnamestor[i]
                tempout = tempout.replace('*','star')
                tempout = tempout.replace('.csv','')
                tempout = tempout.replace('.txt','')
                tempout = tempout + '_trim' + '.csv'
                dfout.to_csv(tempout,sep = '\t',index=False)
            if len(fnameout) > 1:
            	if self.outsumm is not None:
                	dfout_summ = pd.DataFrame({'Velocity':self.outsumm[0,:].T,
                	                          'Wavelength':self.outsumm[1,:].T,
                	                          'Frequency':np.ones(len(
                	                          self.outsumm[1,:])).T,
                	                          '2StDev':self.outsumm[2,:].T})
                	dfout_summ.to_csv(fnameout,sep = '\t',index=False)
        

        
